/* wrap.m     Simple custom wrappers for Portland
 *
 * Copyright (c) 2016 Psellos   http://psellos.com
 * Licensed under the MIT license:
 *     http://www.opensource.org/licenses/mit-license.php
 */
# import <stdlib.h>
# import <objc/runtime.h>
# import <UIKit/UIKit.h>
# import "caml/memory.h"
# import "caml/alloc.h"
# import "caml/callback.h"
# import "wrap.h"

/* Represent raw ObjC object pointers in OCaml as nativeint values.
 */
static value wrapRawObjC(id obj)
{
    CAMLparam0();
    CAMLreturn(caml_copy_nativeint((intnat) obj));
}

static id unwrapRawObjC(value objcval)
{
    return (__bridge id) (void *) Nativeint_val(objcval);
}


/* Some key ObjC objects are wrapped up for access from OCaml. To keep
 * things simple, we enter wrapped ObjC objects into an NSMapTable and
 * never remove them. The table causes the objects to be persistent, and
 * it also lets us map from ObjC objects (pointers) to the OCaml objects
 * that wrap them (values).
 *
 * To stay on good terms with the OCaml GC, the OCaml values must be
 * reachable from a global root. For simplicity, we just make them all
 * global roots individually.
 * 
 * In this simple example these few ObjC objects are never destroyed.
 * In a more dynamic setting, we would need to gain control when the
 * OCaml object is GCed. Then we would remove the entry from the table,
 * which would (most likely) cause it to be deallocated on the ObjC
 * side.
 */
static NSMapTable *g_objc_to_ocaml;

static void init_objc_to_ocaml()
{
    if (g_objc_to_ocaml)
        return;
    NSMapTableOptions kopt =
        NSMapTableStrongMemory | NSMapTableObjectPointerPersonality;
    NSMapTableOptions vopt =
        NSMapTableStrongMemory;
    g_objc_to_ocaml =
        [NSMapTable mapTableWithKeyOptions: kopt valueOptions: vopt];
}


static value find_ocaml_wrapper(id obj) {
    init_objc_to_ocaml();
    NSValue *v = [g_objc_to_ocaml objectForKey: obj];
    if (v == nil)
        return 0;
    return * (value *) [v pointerValue];
}


static value wrapObjC(char *wrapcs, id obj) {
    /* Create a new OCaml object that wraps the given ObjC object.
     * wrapcs gives the name of the closure for creating an OCaml
     * wrapper of the correct type. A closure is registered by each
     * OCaml class that wraps an ObjC class. Each of these OCaml classes
     * inherits from Wrapper.t.
     */
    CAMLparam0();
    CAMLlocal1(val);
    value *valp;
    value *closure;

    closure = (value *) caml_named_value(wrapcs);
    valp = (value *) malloc(sizeof *valp);
    if(closure == NULL || valp == NULL) {
        /* If this happens, things will go very bad from here.
         */
        fprintf(stderr, "wrapObjC failure for %s\n", wrapcs);
        fflush(stderr);
        CAMLreturn(Val_unit); /* Not type safe, should never happen */
    }
    val = caml_callback(*closure, wrapRawObjC(obj));
    *valp = val;
    NSValue *v = [NSValue valueWithPointer: valp];

    /* Enter the wrapping into the table.
     */
    init_objc_to_ocaml();
    [g_objc_to_ocaml setObject: v forKey: obj];
    caml_register_global_root(valp);

    CAMLreturn(val);
}


static id unwrapObjC(value obj) {
    /* Return the ObjC object wrapped by the given OCaml object.
     */
    CAMLparam1(obj);
    CAMLlocal1(objcval);

    objcval =
        caml_callback(
            caml_get_public_method(obj, caml_hash_variant("contents")),
            obj);
    CAMLreturnT(id, unwrapRawObjC(objcval));
}


/* Subclasses of WrapOCaml wrap OCaml objects as ObjC objects.
 */
@implementation WrapOCaml {
    value _contents;
}

- (value) contents
{
    return _contents;
}

- (void) setContents: (value) aValue
{
    _contents = aValue;
}

- (WrapOCaml *) init
{
    /* Create a new OCaml object that is wrapped by self. We match the
     * OCaml class to our own class by name--simple but effective. We
     * then call the closure for creating a wrapped object, which is
     * registered by each OCaml class that can be wrapped in an ObjC
     * class. Such OCaml classes inherit from Wrappee.t.
     */
    const char *classname;
    char closurename[128];
    value *closure;

    if((self = [super init]) != nil) {
        classname = object_getClassName(self);
        if(strlen(classname) + 8 >= sizeof closurename) {
            fprintf(stderr, "[WrapOCaml init]: class name too long: %s\n",
                    classname);
            fflush(stderr);
            return nil;
        }
        strcpy(closurename, classname);
        strcat(closurename, ".wrapped");
        closure = (value *) caml_named_value(closurename);
        if(closure == NULL) {
            fprintf(stderr, "[WrapOCaml init]: lookup failure for %s\n",
                    closurename);
            fflush(stderr);
            return nil;
        }
        _contents = caml_callback(*closure, wrapRawObjC(self));
        caml_register_global_root(&_contents);
    }
    return self;
}

- (void) dealloc
{
    if(_contents != 0) {
        caml_remove_global_root(&_contents);
        _contents = 0;
    }
}

@end

/* OCaml objects accessed from ObjC.
 */

/* An ObjC method name often has ':' in it, which makes it an invalid
 * OCaml method name. Translate ':' in ObjC to '\'' in OCaml. Also, of
 * course, an OCaml method name is a single token.
 */

@implementation PortlandAppDelegate

@dynamic window;

- (UIWindow *) window
{
    CAMLparam0();
    CAMLlocal2(selfval, windowval);

    selfval = [self contents];
    windowval =
        caml_callback(
            caml_get_public_method(selfval, caml_hash_variant("window")),
            selfval);
    CAMLreturnT(UIWindow *, unwrapObjC(windowval));
}


- (void) setWindow: (UIWindow *) aWindow
{
    CAMLparam0();
    CAMLlocal2(selfval, windowval);

    selfval = [self contents];
    windowval = wrapObjC("UiWindow.wrap", (id) aWindow);
    (void)
        caml_callback2(
            caml_get_public_method(selfval, caml_hash_variant("setWindow'")),
            selfval, windowval);
    CAMLreturn0;
}


- (void) applicationDidFinishLaunching: (UIApplication *) anApplication
{
    CAMLparam0();
    CAMLlocal2(selfval, applval);

    selfval = [self contents];
    if((applval = find_ocaml_wrapper(anApplication)) == 0)
        applval = wrapObjC("UiApplication.wrap", (id) anApplication);
    (void)
        caml_callback2(
            caml_get_public_method(
                selfval,
                caml_hash_variant("applicationDidFinishLaunching'")),
            selfval, applval);
    CAMLreturn0;
}

- (void) applicationWillResignActive: (UIApplication *) anApplication
{
    CAMLparam0();
    CAMLlocal2(selfval, applval);

    selfval = [self contents];
    if((applval = find_ocaml_wrapper(anApplication)) == 0)
        applval = wrapObjC("UiApplication.wrap", (id) anApplication);
    (void)
        caml_callback2(
            caml_get_public_method(
                selfval,
                caml_hash_variant("applicationWillResignActive'")),
            selfval, applval);
    CAMLreturn0;
}

- (void) applicationDidBecomeActive: (UIApplication *) anApplication
{
    CAMLparam0();
    CAMLlocal2(selfval, applval);

    selfval = [self contents];
    if((applval = find_ocaml_wrapper(anApplication)) == 0)
        applval = wrapObjC("UiApplication.wrap", (id) anApplication);
    (void)
        caml_callback2(
            caml_get_public_method(
                selfval,
                caml_hash_variant("applicationDidBecomeActive'")),
            selfval, applval);
    CAMLreturn0;
}

@end

@implementation PortlandViewControllerD

@dynamic label;

- (UILabel *) label
{
    CAMLparam0();
    CAMLlocal2(selfval, labelval);

    selfval = [self contents];
    labelval =
        caml_callback(
            caml_get_public_method(selfval, caml_hash_variant("label")),
            selfval);
    CAMLreturnT(UILabel *, unwrapObjC(labelval));
}

- (void) setLabel: (UILabel *) aLabel
{
    CAMLparam0();
    CAMLlocal2(selfval, labelval);

    selfval = [self contents];
    labelval = wrapObjC("UiLabel.wrap", (id) aLabel);
    (void)
        caml_callback2(
            caml_get_public_method(selfval, caml_hash_variant("setLabel'")),
            selfval, labelval);
    CAMLreturn0;
}

- (void) viewController: (UIViewController *) vc
         viewWillAppear: (BOOL) animated
{
    CAMLparam0();
    CAMLlocal2(selfval, vcval);

    selfval = [self contents];
    if ((vcval = find_ocaml_wrapper(vc)) == 0)
        vcval = wrapObjC("PortlandViewController.wrap", (id) vc);
    (void)
        caml_callback3(
            caml_get_public_method(
                selfval,
                caml_hash_variant("viewController'viewWillAppear'")),
            selfval, vcval, Val_bool(!!animated));
    CAMLreturn0;
}


- (void) viewController: (UIViewController *) vc
         viewWillDisappear: (BOOL) animated
{
    CAMLparam0();
    CAMLlocal2(selfval, vcval);

    selfval = [self contents];
    if ((vcval = find_ocaml_wrapper(vc)) == 0)
        vcval = wrapObjC("PortlandViewController.wrap", (id) vc);
    (void)
        caml_callback3(
            caml_get_public_method(
                selfval,
                caml_hash_variant("viewController'viewWillDisappear'")),
            selfval, vcval, Val_bool(!!animated));
    CAMLreturn0;
}


- (BOOL) viewControllerShouldAutorotate: (UIViewController *) vc
{
    CAMLparam0();
    CAMLlocal3(selfval, vcval, resval);

    selfval = [self contents];
    if ((vcval = find_ocaml_wrapper(vc)) == 0)
        vcval = wrapObjC("PortlandViewController.wrap", (id) vc);
    resval =
        caml_callback2(
            caml_get_public_method(
                selfval,
                caml_hash_variant("viewControllerShouldAutorotate'")),
            selfval, vcval);
    CAMLreturnT(BOOL, Bool_val(resval));
}


- (void) timerTick: (NSTimer *) aTimer
{
    CAMLparam0();
    CAMLlocal2(selfval, timerval);


    selfval = [self contents];
    if((timerval = find_ocaml_wrapper(aTimer)) == 0) {
        /* Can't report a timer tick if the timer isn't known.
         */
        CAMLreturn0;
    }
    (void)
        caml_callback2(
            caml_get_public_method(selfval, caml_hash_variant("timerTick'")),
            selfval, timerval);
    CAMLreturn0;
}

@end


@implementation PortlandViewController
/* (Just a thin wrapper that passes methods along to a delegate.)
 */

- (void) viewWillAppear:(BOOL)animated
{
    [self.delegate viewController: self viewWillAppear: animated];
}

- (void) viewWillDisappear:(BOOL)animated
{
    [self.delegate viewController: self viewWillDisappear: animated];
}

- (BOOL) shouldAutorotate
{
    return [self.delegate viewControllerShouldAutorotate: self];
}

@end

value PortlandViewController_delegate(value objcval)
{
    CAMLparam1(objcval);

    PortlandViewController *pvc = unwrapRawObjC(objcval);
    PortlandViewControllerD *delg =  (PortlandViewControllerD *) [pvc delegate];
    CAMLreturn([delg contents]);
}

value PortlandViewController_setDelegate_(value objcval, value delgval)
{
    CAMLparam2(objcval, delgval);  // delgval : PortlandViewControllerD.t
    CAMLlocal1(delgro);            // delgro : "raw" nativeint
    PortlandViewControllerD *delg; // delg: ObjC object

    PortlandViewController *pvc = unwrapRawObjC(objcval);
    delgro = 
        caml_callback(
            caml_get_public_method(delgval, caml_hash_variant("container")),
            delgval);
    delg = unwrapRawObjC(delgro);
    [pvc setDelegate: delg];
    CAMLreturn(Val_unit);
}


/* ObjC objects accessed from OCaml.
 */

/* NSTimer
 */

value NSTimer_scheduledTimer5(value timeInterval, value target, value selector,
                    value userInfo, value repeats)
    /* For simplicity, the target must be wrapped (by an ObjC object)
     * already. It's OK if userInfo isn't wrapped; we treat this as
     * nil.
     */
{
    CAMLparam5(timeInterval, target, selector, userInfo, repeats);
    CAMLlocal2(targro, uiro);

    double ti = Double_val(timeInterval);
    targro = 
        caml_callback(
            caml_get_public_method(target, caml_hash_variant("container")),
            target);
    id targ = unwrapRawObjC(targro);
    if(targ == nil) {
        /* If the target wasn't already wrapped, things will go very bad
         * from here.
         */
        fprintf(stderr, "NSTimer_scheduledTimer5: unwrapped target\n");
        fflush(stderr);
        CAMLreturn(Val_unit); /* Not type safe, should never happen */
    }

    // Caller supplies selector in OCaml form (with '\''). Translate to
    // ObjC form, then translate to selector.
    //
    char *sels = strdup(String_val(selector));
    char *quote = sels;
    while ((quote = strchr(quote, '\'')) != NULL)
        *quote = ':';
    SEL sel = NSSelectorFromString([NSString stringWithUTF8String: sels]);
    free(sels);

    uiro =
        caml_callback(
            caml_get_public_method(userInfo, caml_hash_variant("container")),
            userInfo);
    id ui = unwrapRawObjC(uiro);
    int rep = Bool_val(repeats);

    NSTimer *timer =
        [NSTimer scheduledTimerWithTimeInterval: ti
                 target: targ
                 selector: sel
                 userInfo: ui
                 repeats: rep ];
    CAMLreturn(wrapObjC("NsTimer.wrap", (id) timer));
}


/* UIWindow
 */

value UIWindow_frame(value objcval)
{
    CAMLparam1(objcval);
    CAMLlocal1(b);
    UIWindow *window = unwrapRawObjC(objcval);
    CGRect r = window.frame;
    b = caml_alloc(4, 0);
    Store_field(b, 0, caml_copy_double(r.origin.x));
    Store_field(b, 1, caml_copy_double(r.origin.y));
    Store_field(b, 2, caml_copy_double(r.size.width));
    Store_field(b, 3, caml_copy_double(r.size.height));
    CAMLreturn(b);
}

value UIWindow_rootViewController(value objcval)
{
    /* (The Portland root view controller is a portlandViewController.)
     */
    CAMLparam1(objcval);
    CAMLlocal1(vcob);

    UIWindow *window = unwrapRawObjC(objcval);
    PortlandViewController *vc =
        (PortlandViewController *) [window rootViewController];
    vcob = wrapObjC("PortlandViewController.wrap", vc);
    CAMLreturn(vcob);
}


/* UILabel
 */

value UILabel_setText_(value objcval, value textval)
{
    CAMLparam2(objcval, textval);

    UILabel *label = unwrapRawObjC(objcval);
    NSString *text = [NSString stringWithUTF8String: String_val(textval)];
    [label setText: text];
    CAMLreturn(Val_unit);
}


value UILabel_setBounds_(value objcval, value boundsval)
{
    CAMLparam2(objcval, boundsval);
    CGRect bounds;

    UILabel *label = unwrapRawObjC(objcval);
    bounds.origin.x = Double_val(Field(boundsval, 0));
    bounds.origin.y = Double_val(Field(boundsval, 1));
    bounds.size.width = Double_val(Field(boundsval, 2));
    bounds.size.height = Double_val(Field(boundsval, 3));

    [label setBounds: bounds];
    CAMLreturn(Val_unit);
}

value UILabel_setTransform_(value objcval, value transval)
{
    CAMLparam2(objcval, transval);
    CGAffineTransform trans;

    UILabel *label = unwrapRawObjC(objcval);
    trans.a = Double_val(Field(transval, 0));
    trans.b = Double_val(Field(transval, 1));
    trans.c = Double_val(Field(transval, 2));
    trans.d = Double_val(Field(transval, 3));
    trans.tx = Double_val(Field(transval, 4));
    trans.ty = Double_val(Field(transval, 5));

    [label setTransform: trans];
    CAMLreturn(Val_unit);
}


/* UIDevice
 */

value UIDevice_currentDevice(value unit) {
    CAMLparam1(unit);
    CAMLlocal1(deviceval);

    UIDevice *device = [UIDevice currentDevice];
    if((deviceval = find_ocaml_wrapper(device)) == 0)
        deviceval = wrapObjC("UiDevice.wrap", (id) device);
    CAMLreturn(deviceval);
}

value UIDevice_beginGeneratingDeviceOrientationNotifications(value objcval)
{
    CAMLparam1(objcval);

    UIDevice *device = unwrapRawObjC(objcval);
    [device beginGeneratingDeviceOrientationNotifications];
    CAMLreturn(Val_unit);
}

value UIDevice_orientation(value objcval)
{
    CAMLparam1(objcval);

    UIDevice *device = unwrapRawObjC(objcval);
    UIDeviceOrientation orient = [device orientation];
    if(     orient < UIDeviceOrientationUnknown ||
            orient > UIDeviceOrientationFaceDown)
        orient = UIDeviceOrientationUnknown;
    CAMLreturn(Val_int((int) orient));
}


/* UIApplication
 */

value UIApplication_delegate(value objcval)
{
    /* (The Portland application delegate is coded in OCaml, so it
     * appears in ObjC as a subclass of WrapOCaml.)
     */
    CAMLparam1(objcval);
    UIApplication *appl = unwrapRawObjC(objcval);
    WrapOCaml *deleg = (WrapOCaml *) [appl delegate];
    CAMLreturn([deleg contents]);
}


value UIApplication_sharedApplication(value unitval)
{
    CAMLparam1(unitval);
    CAMLlocal1(applval);
    UIApplication *application = [UIApplication sharedApplication];
    if((applval = find_ocaml_wrapper(application)) == 0)
        applval = wrapObjC("UiApplication.wrap", (id) application);
    CAMLreturn(applval);
}
